// *****
// * PROJECT:		Memory utilities
// * FILENAME: 		MemDbg.cpp
// * AUTHOR:		Jeff Ishaq 05/12/00 jishaq@earthlink.net
// * 
// * DESCRIPTION:	Implementation for memory testing tool.
// *
// * NOTES:         To enable leak detection, you want to #include "MemDbgMacros.h"
// *                at a very high level.  I suggest using the prefix file to do this,
// *                to ensure MemHandleNew() and such get overridden properly (see
// *                build.h for more on this).  Only the files in whose scope 
// *                MemDbgMacros.h exists will be checked for memory leaks.
// *
// *                
// * COPYRIGHT:  	With the exception of this comment block, you are free to modify
// *				and/or to reuse this source code as you see fit.  For the sake
// *				of documentation, I ask that you leave this base comment block
// *                intact, and I encourage you to extend it as necesary.
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *    JeffI   04/27/01    Tidied this up in preparation for publication
// *****

// Only compile in this code if we need to:
#ifdef CHECK_FOR_MEMORY_LEAKS

// MemDbg.h is already included in the prefix file 'build.h'...
#include "MyPilot.h"

// Turn off our remapping macros here!!!  If we do not do this, we will have infinite
// recursion in _DbgMemHandleNew(), among other bad things.  We switch these macros back
// on at the bottom of this file.  This only has to happen here, it's a one-time 'trick':
#undef MemHandleNew
#undef MemPtrNew
#undef new

// *****
// * FUNCTION: 		operator new
// * 
// * DESCRIPTION:	Wrapper for new, using the 'placement new' facility of C++ to allow
// *				arbitrary arguments to be passed in.  Pretty sneaky.
// *
// * ARGUMENTS:		size		-	size, in bytes, of chunk to allocate
// *				pFileName	-	name of file in which allocation took place
// *				lineNum		-	numer of the line on which allocation took place
// *
// * RETURN:		(VoidPtr) Ptr to allocated memory
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	08/07/00	Initial Implementation
// *****
VoidPtr operator new (UInt32 size, CharPtr pFileName, ULong lineNum )
{
	return _DbgMemPtrNew( size, pFileName, lineNum);
}
	
	
// *****
// * FUNCTION: 		operator new[]
// * 
// * DESCRIPTION:	Wrapper for new[], using the 'placement new' facility of C++ to allow
// *				arbitrary arguments to be passed in.  Pretty sneaky.
// *
// * ARGUMENTS:		size		-	size, in bytes, of chunk to allocate
// *				pFileName	-	name of file in which allocation took place
// *				lineNum		-	numer of the line on which allocation took place
// *
// * RETURN:		(VoidPtr) Ptr to allocated memory
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	08/07/00	Initial Implementation
// *****
VoidPtr operator new [] ( UInt32 size, CharPtr pFileName, ULong lineNum )
{
	return _DbgMemPtrNew( size, pFileName, lineNum );
}

// *****
// * FUNCTION: 		_DbgMemHandleNew
// * 
// * DESCRIPTION:	Wrapper for MemHandleNew, called automatically by #define in
// *				MemDbgMacros.h file that remaps MemHandleNew() to _DbgMemHandleNew().
// *
// * ARGUMENTS:		size		-	size, in bytes, of chunk to allocate
// *				pFileName	-	name of file in which allocation took place
// *				lineNum		-	numer of the line on which allocation took place
// *
// * RETURN:		(VoidHand) Handle to allocated memory
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****
VoidHand _DbgMemHandleNew( UInt size, CharPtr pFileName, ULong lineNum )
{
	VoidHand h = MemHandleNew( size );

	// Log this handle allocation:
	AddDbgEntry( pFileName, size, lineNum, h );
	
	// Return the handle to caller now:
	return h;
}	

// *****
// * FUNCTION: 		_DbgMemPtrNew
// * 
// * DESCRIPTION:	Wrapper for MemPtrNew, called automatically by #define in
// *				MemDbgMacros.h file that remaps MemPtrNew() to _DbgMemPtrNew().
// *
// * ARGUMENTS:		size		-	size, in bytes, of chunk to allocate
// *				pFileName	-	name of file in which allocation took place
// *				lineNum		-	numer of the line on which allocation took place
// *
// * RETURN:		(VoidPtr) Ptr to allocated memory
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****
VoidPtr _DbgMemPtrNew( UInt size, CharPtr pFileName, ULong lineNum )
{
	VoidPtr p = MemPtrNew( size );
	
	// Log this ptr allocation:
	AddDbgEntry( pFileName, size, lineNum, p );
	
	// Return to caller:
	return p;
}	

	
// *****
// * FUNCTION: 		_DbgMemHandleFree
// * 
// * DESCRIPTION:	Wrapper for MemHandleFree, called automatically by remapping the
// *                sysTrapMemHandleFree trap (see HookMemMgr()).
// *
// * ARGUMENTS:		h		    -   Handle to free
// *
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****
Err _DbgMemHandleFree( VoidHand h )
{
	// See if this entry exists in the linked list.  If not, we must not have allocated it,
	// so we just free it and carry on
	UInt index = FindDbgEntry (h);
	if ( index == dmMaxRecordIndex )	// if not found, deallocate and return
	{
		
		return gpOldMemHandleFree( h );
	}
	
	// Lock down the node so we can look at it:
	VoidHand hNode = DmQueryRecord( MemDbgRef, index );
	NodePtr  pNode = (NodePtr) MemHandleLock( hNode );

	// Deallocate the actual memory chunk:
	Err err = gpOldMemHandleFree( pNode->hAddr );
	
	// Unlock the record data and delete it from the DB:
	MemHandleUnlock( hNode ); pNode = NULL;
	KillRecord( index );

	return err;
}

// *****
// * FUNCTION: 		_DbgMemPtrFree
// * 
// * DESCRIPTION:	Wrapper for MemPtrFree, called automatically by remapping the
// *                sysTrapMemChunkFree trap (see HookMemMgr()).
// *
// *                NOTE:  In PalmOS SDK, MemPtrFree calls through to MemChunkFree!
// *
// * ARGUMENTS:		p		    -   Ptr to free
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****
Err _DbgMemChunkFree( VoidPtr p )
{
	// See if this ptr exists in our database:
	UInt index = FindDbgEntry( p );

	// If we found it, remove the record from our DB:
	if ( index != dmMaxRecordIndex )
		KillRecord( index );
		
	// In any case, we delete the ptr as requested:
	return gpOldMemChunkFree( p );
}

// *****
// * FUNCTION: 		AddDbgEntry( ..., VoidPtr overload)
// * 
// * DESCRIPTION:	Add a pointer address into our database
// *
// * ARGUMENTS:		pFileName	-	name of file in which allocation took place
// *				size		-	size, in bytes, of chunk to allocate
// *                lineNum		-	numer of the line on which allocation took place
// *				pAddr		-	the pointer address (primary key)
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****
void AddDbgEntry(CharPtr pFileName, ULong size, ULong lineNum, VoidPtr pAddr )
{
	Node theNode;
	StrNCopy( theNode.fileName, pFileName, MemDbgFilenameSize );
	theNode.size 		= size;
	theNode.lineNum 	= lineNum;
	theNode.type		= PointerType;
	theNode.pAddr		= pAddr;
	_AddDbgEntry( &theNode );			// Call through to helper fcn to do the work
}

// *****
// * FUNCTION: 		AddDbgEntry( ..., VoidHand overload )
// * 
// * DESCRIPTION:	Add a handle into our database
// *
// * ARGUMENTS:		size		-	size, in bytes, of chunk to allocate
// *				pFileName	-	name of file in which allocation took place
// *				lineNum		-	numer of the line on which allocation took place
// *				hAddr		-	the handle address (primary key)
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****
void AddDbgEntry(CharPtr pFileName, ULong size, ULong lineNum, VoidHand hAddr )
{
	Node theNode;
	StrNCopy( theNode.fileName, pFileName, MemDbgFilenameSize );
	theNode.size 		= size;
	theNode.lineNum 	= lineNum;
	theNode.type		= HandleType;
	theNode.hAddr		= hAddr;
	_AddDbgEntry( &theNode );			// Call through to helper fcn to do the work
}

// *****
// * FUNCTION: 		_AddDbgEntry()
// * 
// * DESCRIPTION:	Actually inserts a NodePtr into the database
// *
// * ARGUMENTS:		pNodeToAdd		-		The Node to append into the DB
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****
void _AddDbgEntry( NodePtr pNodeToAdd )
{
	// If the DB isn't open, generate a fresh one and open it:
	if ( !MemDbgRef )
	{
		MakeMemDbgDB();
	}
	
	// Pack up the record:
	UInt index = dmMaxRecordIndex;		// Always add to the end of DB
	VoidHand recH  = DmNewRecord(	MemDbgRef,
									&index,
									sizeof( Node ));
	ErrFatalDisplayIf( recH == 0, "_AddDbgEntry: Unable to create a new record.");

	NodePtr recP = (NodePtr) MemHandleLock( recH );
	Err err = DmWrite( recP, 0, pNodeToAdd, sizeof( Node ));
	ErrFatalDisplayIf( err != 0, "_AddDbgEntry: Unable to DmWrite into the new record!");
	MemHandleUnlock( recH );
	
	DmReleaseRecord(	MemDbgRef,
						index,
						false );		// Don't set dirty bit (we don't care)
						
}

// *****
// * FUNCTION: 		_FindDbgEntry()
// * 
// * DESCRIPTION:	Given an address, locate in the
// *				database
// *
// * ARGUMENTS:		addr		-		key to match
// *				type		-		type of allocation (ptr or handle?)
// *
// * RETURNS:		Index into the DB of the found record, or dmMaxRecordIndex if not
// *				found.
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****
UInt _FindDbgEntry( VoidPtr addr, eAllocationType type )
{
	// If we haven't created the DB yet, don't
	if ( !MemDbgDBID )
		return dmMaxRecordIndex;
		
	// Run through the DB to see if we find hAddr in it
	UInt numRecords = DmNumRecords( MemDbgRef );
	UInt i = 0;
	for ( ; i < numRecords; i++ )
	{	
		//Query the current record:
		VoidHand recH = DmQueryRecord( MemDbgRef, i );
		ErrFatalDisplayIf( recH == 0, "FindDbgEntry: Unable to query record.");
		NodePtr recP = (NodePtr) MemHandleLock( recH );
		
		// See if we've got a match.  If so, break out of for{}
		if ( (recP->type == type) && (recP->pAddr == addr) )
			break;

		// If not, unlock it and go to the next one
		MemHandleUnlock( recH ); recP = NULL;
	}
	
	// Return dmMaxRecordIndex on failure, otherwise return the idx
	if ( i == numRecords )
		return dmMaxRecordIndex;
	else
		return i;
}

// *****
// * FUNCTION: 		FindDbgEntry( VoidHand )
// * 
// * DESCRIPTION:	Given a handle, locate either a pointer or a handle in the
// *				database.  Calls through to helper function
// *
// * ARGUMENTS:		hAddr		-		key to match
// *
// * RETURNS:		Index into the DB of the found record, or dmMaxRecordIndex if not
// *				found.
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****
UInt FindDbgEntry( VoidHand hAddr )
{
	return (_FindDbgEntry( (VoidPtr)hAddr, HandleType ));
} 

// *****
// * FUNCTION: 		FindDbgEntry( VoidPtr )
// * 
// * DESCRIPTION:	Given a pointer, locate in the
// *				database.  Calls through to helper function
// *
// * ARGUMENTS:		pAddr		-		key to match
// *
// * RETURNS:		Index into the DB of the found record, or dmMaxRecordIndex if not
// *				found.
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****
UInt FindDbgEntry( VoidPtr pAddr )
{
	return (_FindDbgEntry( pAddr, PointerType ));
} 

// *****
// * FUNCTION: 		RemoveDbgEntry( VoidHand )
// * 
// * DESCRIPTION:	Given a handle, remove its record from the DB.  Does NOT deallocate
// *				it!  Calls through to helper function
// *
// * ARGUMENTS:		hAddr		-		key to match
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****	
void RemoveDbgEntry(VoidHand hAddr )
{
	_RemoveDbgEntry((VoidPtr)hAddr, HandleType );
}

// *****
// * FUNCTION: 		RemoveDbgEntry( VoidPtr )
// * 
// * DESCRIPTION:	Given a pointer, remove its record from the DB.  Does NOT deallocate
// *				it!  Calls through to helper function
// *
// * ARGUMENTS:		pAddr		-		key to match
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****	
void RemoveDbgEntry(VoidPtr pAddr )
{
	_RemoveDbgEntry(pAddr, PointerType );
}

// *****
// * FUNCTION: 		_RemoveDbgEntry()
// * 
// * DESCRIPTION:	Given an address and a type, remove the record from the DB.
// *
// * ARGUMENTS:		addr		-		key to match.  Can be ptr or handle
// *				type		-		type, also a match key.
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****	
void _RemoveDbgEntry( VoidPtr addr, eAllocationType type )
{
	// Find the record:
	UInt i = _FindDbgEntry( addr, type );
	ErrFatalDisplayIf( i == dmMaxRecordIndex, "_RemoveDbgEntry: Unable to find the address!");

	// KILL the record in question:
	KillRecord( i );
}

// *****
// * FUNCTION: 		HookMemMgr()
// * 
// * DESCRIPTION:	Patch the traps of the deallocation functions.  This is a total
// *				hack, only use this for debugging sessions!  I had originally just
// *				#defined MemHandleFree _DbgMemHandleFree, but this did not work
// *				very well because I allocated a lot of form fields that I let the OS clean
// *				up automatically on frmClose.  A #define can't remap already-compiled
// *				OS code calls, but redefining a trap sure can.  These were showing up
// *				as leaks because even though the OS deallocated the form fields I had
// *				allocated, it never went through _DbgMemHandleFree so our DbgDB was not
// *				balanced.
// *
// * POSTCONDITION:	After you're done testing, be sure to restore the traps to their
// *				(assumed) original values by calling CloseMemDbgDB()!
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****	
void HookMemMgr()
{
	gpOldMemHandleFree = (myMemHandleFreePtr)SysGetTrapAddress( sysTrapMemHandleFree );
	SysSetTrapAddress( sysTrapMemHandleFree, _DbgMemHandleFree );
	
	gpOldMemChunkFree = (myMemChunkFreePtr)SysGetTrapAddress( sysTrapMemChunkFree );
	SysSetTrapAddress( sysTrapMemChunkFree, _DbgMemChunkFree );
}

// *****
// * FUNCTION: 		KillRecord()
// * 
// * DESCRIPTION:	Simply delete a record out of the opened DB
// *
// * ARGUMENTS:		i		-		index of record to remove
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****	
void KillRecord( UInt i )
{
	Err err = DmRemoveRecord( MemDbgRef, i );
	ErrFatalDisplayIf( err != 0, "KillRecord: Error removing record from DB.");
}
	
// *****
// * FUNCTION: 		CheckForLeaks()
// * 
// * DESCRIPTION:	This will report on any remaining Nodes left in the DB.
// *				If this is called after the app is almost finished, this
// *				will show all the allocations that were never allocated.
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****	
void CheckForLeaks()
{
	// Go through all records in the db:
	UInt i = DmNumRecords( MemDbgRef );
	for ( UInt j = 0; j < i; j++ )
	{
		// Display diagnostic info about this record.  We don't delete them.
		VoidHand hRec = DmQueryRecord( MemDbgRef, j );
		NodePtr  pRec = (NodePtr) MemHandleLock (hRec);

		// Good old StrPrintF was not working for this, so we do it manually:
		char output[255];
		char line[7];
		char size[7];
		StrIToA( line, pRec->lineNum );
		StrIToA( size, pRec->size );
		StrCopy( output, pRec->fileName );
		StrCat( output, ", line " );
		StrCat( output, line );
		StrCat( output, ", size ");
		StrCat( output, size );
		StrCat( output, ":  ");
		StrCat( output, "leaked!" );

		// Display the debug box here:
		ErrNonFatalDisplay( output ); 
		
		// Ptr to the actual buffer, for ease of debugging:
		CharPtr pBuf;
		Boolean bAttemptToLock = 0;
		if ( pRec->type == HandleType )
		{
			if ( bAttemptToLock )
				CharPtr pBuf = (CharPtr) MemHandleLock( pRec->hAddr );
		}
		else
			pBuf = (CharPtr)pRec->pAddr;


		// Now unlock the buffer ptr, we no longer need it
		if ( pRec->type == HandleType )
		{
			if ( bAttemptToLock )
				MemHandleUnlock( pRec->hAddr );
		}
		
		pBuf = NULL;
		MemHandleUnlock( hRec );

	}
}

// *****
// * FUNCTION: 		UnhookMemMgr()
// * 
// * DESCRIPTION:	Call this when your app is about to quit.  It will close the DbgDB,
// *                and attempt to restore the remapped traps.
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****	
void UnhookMemMgr()
{
	if (!MemDbgRef)
	{
		ErrNonFatalDisplay("The Dbg DB's ref num does not exist, so I can't close it.  Nonfatal.");
		return;
	}
	
	if ( gpOldMemHandleFree )
		SysSetTrapAddress( sysTrapMemHandleFree, gpOldMemHandleFree );	
		
	if ( gpOldMemChunkFree )
		SysSetTrapAddress( sysTrapMemChunkFree, gpOldMemChunkFree );

    // Note:  We want this call to happen --AFTER-- we have un-patched the two traps above.
    // This is because DmMgr ends up calling one/both of the patched functions.  This 
    // prevents us from crashing when we close the database (which calls MemHandleFree,
    // which is patched to _DbgMemHandleFree(), which in turn calls AddDbgEntry, which in
    // turn trys to write in the DB which is presently being closed!)
    // crash if
	DmCloseDatabase( MemDbgRef );
	MemDbgRef = NULL;
	MemDbgDBID = NULL;
	
}
		
	
// *****
// * FUNCTION: 		MakeMemDbgDB()
// * 
// * DESCRIPTION:	This will find, clobber, and create a new Debug db, 
// *				leaving it opened.
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *	JeffI	05/12/00	Initial Implementation
// *****	
void MakeMemDbgDB()
{
	// See if our DB exists yet.  If so, delete it, and create a fresh one.
	LocalID dbID;
	dbID = DmFindDatabase(		MemDbgDBCardNum,
								MemDbgDBName );
	Err err;
	if ( dbID )
	{
		err = DmDeleteDatabase( 	MemDbgDBCardNum,
										dbID );
		ErrFatalDisplayIf( err, "MakeMemDbgDB: Unable to delete existing MemDbgDB!");
	}
	
	err = DmCreateDatabase(	MemDbgDBCardNum,
							MemDbgDBName,
							MemDbgCreatorID,
							MemDbgType,
							false );	// resource DB = falase
	if ( err )
	{
		ErrFatalDisplay("MakeMemDbgDB: Unable to create the Preferences database.");
	}
		
	dbID = DmFindDatabase(	MemDbgDBCardNum,
							MemDbgDBName );
	
	// Open, and leave it opened.
	DmOpenRef dbRef;
	dbRef = DmOpenDatabase(	MemDbgDBCardNum,
							dbID,
							dmModeReadWrite );
	
	MemDbgDBID 	= dbID;
	MemDbgRef	= dbRef; 
}	

// This will 'turn back on' the macro remappings that we turned off at the top of this file
// to prevent infinite recursion in _DbgMemHandleNew() and similar functions
#include "MemDbgMacros.h"
#endif  // CHECK_FOR_MEMORY_LEAKS