// *****
// * PROJECT:		Memory utilities
// * FILENAME: 		Starter.cpp
// * AUTHOR:		Jeff Ishaq 05/12/00 jishaq@earthlink.net
// * 
// * DESCRIPTION:	Example illustrating Memory Debugging with these tools
// *
// * NOTES:         Leak detection in this example code is 'turned on' by
// *                two things:
// *                    1) The inclusion of MemDbgMacros.h, which serves to
// *                       remap MemPtrNew(), MemHandleNew(), and operator new.
// *                       This handles remapping of the ALLOCATION functions.
// *
// *                    2) Calling 'HookMemMgr' and subsequently calling UnhookMemMgr()
// *                       at the beginning/end of application execution.  This serves
// *                       to remap MemHandleFree and MemPtrFree,  which handles remapping
// *                       DEALLOCATION functions
// *
// *                In a 'real' program, #1 would take place in every source code module
// *                that performs allocations.  An easy way to make this happen is by using
// *                the prefix file -- kind of a 'globals' are for preprocessing macros such
// *                as #define and #include.  See build.h for information on this, and an alternative.
// *
// *                #2 would still remain in the PilotMain, just as it appears here.  You want
// *                to hook the PalmOS Memory Manager right off the bat.  And you want to unhook
// *                it as the very last thing you do before your program exits.  This ensures paired
// *                additions and removals of allocation entries to the MemDbgDB.
// *
// *                In Codewarrior, you can do Project : Preprocess and take a look at the post-macro-
// *                substitution code to get a little better picture of what's going on.
// *
// *                
// * COPYRIGHT:  	With the exception of this comment block, you are free to modify
// *				and/or to reuse this source code as you see fit.  For the sake
// *				of documentation, I ask that you leave this base comment block
// *                intact, and I encourage you to extend it as necesary.
// *
// * REVISION HISTORY:
// *	NAME	DATE		DESCRIPTION
// *    -------------------------------------------------------------------------------
// *    JeffI   04/27/01    Tidied this up in preparation for publication
// *****

#include "MyPilot.h"

class LeakedClass	// The world's most useless class
{
};


DWord PilotMain( Word wCmd, Ptr unusedP, Word unusedW2)
{
    // To keep our demo app simple, we only handle the 'normal launch' code.  Failure
    // to do this without a proper launch code handling loop can cause problems, for
    // example when we get a reset launch code.
    if ( sysAppLaunchCmdNormalLaunch != wCmd )
        return 0;
        
    #ifdef CHECK_FOR_MEMORY_LEAKS // Switch lives in build.h prefix file
        // Patch the traps which the PalmOS Memory Manager calls to deallocate
        // stuff.  The functions that allocate stuff are remapped by virtue of
        // including MemDbgMacros.h:
    	HookMemMgr();
    #endif // CHECK_FOR_MEMORY_LEAKS
	
	// Calling the 'new' operator will actually call thru to our 'operator new()'
	// overload, and in turn, thru to _DbgMemPtrNew.  This happens because we have
	// a macro replacement (defined in MemDbgMacros.h) for new:
	LeakedClass* pClass = new LeakedClass; pClass = NULL;	// Oops! 	(line 69)

    // This will call thru to _DbgMemPtrNew() because of the macro mapping in MemDbgMacros.h:
	MemPtrNew(57);											// Oh dear! (line 72)
	
	// This will call thru to _DbgMemHandleNew() because of the macro mapping in MemDbgMacros.h:
	VoidHand hLeak = MemHandleNew(3000);					// Ouch!	(line 75)

    // This is just an example of an allocation/deallocation
    // that does not leak.  Note that an entry in our MemDbgDB
    // will be created as part of MemHandleNew(), and this entry
    // will then be removed as part of MemHandleFree().
    VoidHand hNotLeaked = MemHandleNew(60);                 
    MemHandleFree( hNotLeaked ); hNotLeaked = NULL;
    
    #ifdef CHECK_FOR_MEMORY_LEAKS  // Switch lives in build.h prefix file
        // Ok, our application is done and ready to close.  We want to run thru the
        // MemDbgDB and note any allocation records that still remain:  it is very likely
        // that these records are memory leaks!
    	CheckForLeaks();
    	
    	// Attempt to revert the traps which we've patched at the top of this function via
    	// HookMemMgr().  If we don't do this, anybody (the OS, the next app we run, Netlib,
    	// etc) will 
    	UnhookMemMgr();
	#endif // CHECK_FOR_MEMORY_LEAKS

	return 0;
}
